#!/usr/bin/python

#############################################
#Author:        Damiano Barboni <barboni@meeo.it>
#Version:       
#Description:   Perform parallel read-only requests to RASDAMAN or PETASCOPE
#Usage:         ./test_parallel_request.py APP N
#                        APP can be
#                            "ras" for RASDAMAN                  
#                            "wcs" for PETASCOPE wcs2.0
#                            "wcps" for PETASCOPE wcps abstract syntax
#                        N is the number of requests generated

#Changelog:     Tue Apr 17 12:35:20 CEST 2012
#               Submit requests also to wcps (both xml or abstract syntax)
#############################################

import httplib
import urllib
import sys
import threading
import time
import commands
import traceback

####################################################
######## global variables definition ###############
####################################################

WCS_host    = "rasdaman"
WCS_port    = 8080
WCS_url     = "/petascope/wcs2?"
WCS_service = "WCS"
WCS_version = "2.0.0"

WCPS_url     = "/petascope/wcps"

CoverageId     = "NIR"

Output_File_Path = "/home/damiano"

#################################
#global variables for subset/trim
_min_x = 0 
_max_x = 200
_min_y = 0
_max_y = 200
###############################
######################################################

class GetCoverage():
    #########################################
    ############## INIT CLASS ###############
    #########################################
    def __init__(self, **kw):
        #initlog
        self.Request = "GetCoverage"
        self.service = WCS_service
        self.url     = WCS_url
        self.subset  = ""
        
        #set coverage id
        self.CoverageId = kw['CoverageId']
          
        #set wcs version
        self.version = WCS_version
        
    #########################################
    ############# SET SUBSET METHOD #########
    #########################################
    def setSubset(self, axis, minVal, maxVal):
        #if the axis is different than x or y log a warning message becuse the axis was not tested 
        if axis.lower()!="x" and axis.lower()!="y" and axis.lower()!="t":
            self.LOG.warning("%s the axis %s is different than x-y-t and was not tested" %(self.Request, axis))
            #append new subset to self.subset string
        if axis.lower()=="t":
            self.subset += "subset%s=%s(%s,%s)&" %(axis.upper(), axis.lower(), minVal, maxVal)
        else:#axis.lower()=="x" OR axis.lower()=="y"
            self.subset += "subset%s=%s(%d,%d)&" %(axis.upper(), axis.lower(), minVal, maxVal)
        
        return self.subset
    
    #########################################
    ############ Get data METHOD ############
    #########################################
    
    def getData(self, **kw):
        #get host coverage id
        host = kw['host']
        port = kw['port']
        
        #create string
        request_string = WCS_url + "service=WCS" + "&Request=" + self.Request + "&version=" + self.version + "&CoverageId=" + self.CoverageId + "&" + self.subset
        
        conn = WCSRequestGet(host, port)
        xml = conn.execute(request_string)
        
        return xml

class WCPSClient():
    #########################################
    ############## INIT CLASS ###############
    #########################################
    def __init__(self, **kw):
        
        #init subset
        try:
            self.subset = kw['subset']
        except:
            self.subset = {'x' : None,
                           'y' : None,
                           't' : None}
        
        self.output = "csv"        
        self.CoverageId = kw['CoverageId']
            
    #############################################
    ############# WCPS QUERY GENERATION #########
    #############################################
    def queryGen(self):
        self.query = ""
        try:
            #set coverage id
            self.query += 'for data in (%s)' %(self.CoverageId)
            
            #set encode
            self.query += ' return encode( data'
            
            ######################
            #         TRIM       #
            ######################
            subset_query = ""
            try:
                if self.subset['x']:
                    subset_query += ' x:"CRS:1"(%s:%s),' %(str(self.subset['x']['min']), str(self.subset['x']['max']))
            except:
                None
            try:
                if self.subset['y']:
                    subset_query += ' y:"CRS:1"(%s:%s)' %(str(self.subset['y']['min']), str(self.subset['y']['max']))
            except:
                None
            
            #if subset is define add the "trim" operator to the WCPS query
            if subset_query:
                self.query = self.query.replace("encode(", "encode( trim(")
                subset_query = ', { ' + subset_query + ' })'
            
            self.query += subset_query
            
            #set output format
            self.query += ', "%s")' %(self.output)
       
        except:
            None
            
        return self.query
        
    #############################################
    ############# WCPS QUERY GENERATION #########
    #############################################
    def xmlGen(self):
        #XML is not currently implemented
        return ""
    
    ##################################################
    ############ Post request data METHOD ############
    ##################################################
    
    def postRequest(self, **kw):
        #get host coverage id
        host = kw['host']
        port = kw['port']
        url  = WCPS_url
        WCPS_type = kw['type']
           
        #get ProcessCoverage request
        if WCPS_type == "query":
            WCPS_request = self.queryGen()
            
        elif WCPS_type == "xml":
            WCPS_request = self.xmlGen()
        
        #submit request
        try:
            conn              = WCPSRequestPost(host, port, url)
            ProcessedCoverage = conn.execute(WCPS_type, WCPS_request)
        except:
            print ("Some errors occur retrivin ProcessedCoverage data.\n%s" %(traceback.format_exc()))
        
        return ProcessedCoverage
         
#implement http reguest to WCS server
class WCSRequestGet():
    def __init__(self, host, port = WCS_port):
        self.host = host
        self.port = port
        
    def execute(self, wcs_request_string):
        #init connection
        conn = httplib.HTTPConnection(self.host+":"+str(self.port))
        conn.request("GET", wcs_request_string)
        r = conn.getresponse()
        
        #check if response status is 200
        r_status = r.status
        if 200 <= r_status <= 299:
            xml = r.read()
        else:
            xml = "ERROR %d" %(r_status)
            print ("Unable to get data from WCS server. Bad response status N %d returned by WCS server" %(r_status))
        
        conn.close()                                
        return xml

#implement http reguest POST to WCPS server
class WCPSRequestPost():
    def __init__(self, host, port, url=WCPS_url):
        self.host = host
        self.port = port
        self.url  = url

    def execute(self, parameter, request_string):
        
        #TAR.REP Request
        headers = {"Content-type": "application/x-www-form-urlencoded",
                   "Accept": "text/plain"}        

        params = urllib.urlencode({parameter : request_string}) 
        #established connection
        conn = httplib.HTTPConnection(self.host, self.port)        
        conn.request("POST", self.url, params, headers)

        #get data
        response = conn.getresponse()
        processedCoverage = response.read()
        #close connection
        conn.close()
                    
        return processedCoverage
        
class requestThread(threading.Thread):
    def __init__(self,application, i):
        threading.Thread.__init__(self)
        self.application = application
        self.th_number   = i
    
    #run method       
    def run(self):
        start = time.time()
        print "Thread %d STARTED" %(self.th_number)
        if self.application == "ras":
            #request to rasdaman
            rasdaman_request = 'rasql -q "select csv(n[%d:%d,%d:%d]) from NIR as n" --out string' %(_min_x, _max_x, _min_y, _max_y)
            data_Out      =  str(commands.getstatusoutput(rasdaman_request))
            
            filename = "%s/%s_output_%d.xml" %(Output_File_Path, self.application, self.th_number)
            
        elif self.application == "wcs":
            #request to petascope
            obj = GetCoverage(CoverageId=CoverageId)
            #set subset
            obj.setSubset("x", _min_x, _max_x)
            obj.setSubset("y", _min_y, _max_y)
    
            #get xml response
            data_Out = obj.getData(host=WCS_host, port=WCS_port)
            
            filename = "%s/%s_output_%d.xml" %(Output_File_Path, self.application, self.th_number)
        
        elif self.application == "wcps":
            #request to petascope
            obj = WCPSClient(CoverageId=CoverageId)
            
            #set subset
            obj.subset['x'] = {'min':_min_x,     'max':_max_x,     'EPSG':1}
            obj.subset['y'] = {'min':_min_y,     'max':_max_y,     'EPSG':1}
    
            #get xml response
            data_Out = obj.postRequest(host=WCS_host, port=WCS_port, type="query")
            
            filename = "%s/%s_output_%d.txt" %(Output_File_Path, self.application, self.th_number)
            
        #write output to file
        out_file = open(filename,"w")
        out_file.write(data_Out)
        out_file.close()
            
        print "Thread %d ENDED in %f seconds" %(self.th_number,time.time()-start)
        sys.exit(0)
    
if __name__ == '__main__':
    try:
        application = sys.argv[1]
        req_number = int(sys.argv[2],10)
        if (application != "ras" or application != "wcs" or application != "wcps") and req_number < 1:
            raise
    except:
        print "ERROR bad arguments. USAGE:\n\tpython %s APP N" %(sys.argv[0])
        print '\t\tAPP can be:'
        print '\t\t\t\t"ras" for RASDAMAN,'
        print '\t\t\t\t"wcs" for PETASCOPE wcs2.0,'
        print '\t\t\t\t"wcps" for PETASCOPE wcps abstract syntax\n'
        print '\t\tN is the number of requests generated'
        sys.exit(65)
    
    #thread generation
    thread_list = []
    for i in range(0, req_number):
        thread_list.append(requestThread(application, i))
        #thread execution
        thread_list[i].start()
        
    #Wait for the thread_list tasks to finish
    for i in range(0,len(thread_list)):
        thread_list[i].join()
        
    print "COMPLETE!"
    
    sys.exit(0)
    
